
#import "Controller.h"
#import "BibliographicFile.h"
#import "BibTeXObject.h"
#import "Preferences.h"
#import "MyCell.h"
#import <sys/time.h>


static	char  *types[] = 
	{	"bib", "bib~",
		NULL
	};


struct rusage stime, etime;
static	int	sMSec, eMSec;


// temporary stream
static NXStream	*localStream = NULL;

#define NEWLINE		fwrite( "\n", sizeof(char), 1, fp )

#define	OUTPUT(obj)	\
	{	r = [obj range];\
		fwrite( "\n", sizeof(char), 1, fp );\
		fwrite( &theBuffer[ r.start ], sizeof(char), r.length, fp );\
	}

//	special entryNames
static	int	STRING = -1;
static	int	PREAMBLE = -1;


@implementation BibliographicFile

	/*************	class methods	***********/

+ (BOOL) canLoadFromFile:(const char *)fName
{
	int	i;
	char *extension;

	// is fullName valid ??
	extension = rindex( fName, '.' );

	if( extension==0 )	return NO;

	extension++;
	for( i=0; types[i]; i++ )
		if( strcmp( extension, types[i] )==0 ) return YES;

	return NO;
}


+ (char **) fileTypes
{
	return types;
}

	/*************	end of class methods	***********/


- (int) compare:(locType)a with:(locType)b
{
	int	len =  (a.length < b.length) ?
				a.length : b.length;
	return
	  ( strncasecmp( &theBuffer[a.start], &theBuffer[b.start], len) > 0 );

}


// inserts 'theObject' into the already sorted list 'listOfObjects'
// returns the index of the insertion point
- (int)insertObject:theObject
{
	int	j = [listOfObjects count] - 1;
	locType keyRange = [theObject key];

	NXGetMemoryBuffer(memStream, &theBuffer, &length, &maxLength);
	while( j>=0 ){
		locType keyRangeL = [[listOfObjects objectAt:j] key];

		if( [self compare:keyRangeL with:keyRange] > 0 ) j--;
		else break;
	}
		
	[listOfObjects insertObject:theObject at:j+1];

	return (j+1);
}

// returns the new position of obj in the list
- (int) updateObjectListFor:obj
{
	[listOfObjects removeObject:obj];	// key may have changed
	return [self insertObject:obj];
}



//	comb sort
- (void) sortArray:(List *)aList;
{
	int	gap, switches, top, i, j;
	int	numElements = [aList count];

	gap = numElements;
	do{
		gap = (int)( (float)gap/1.3 );
		switch( gap ){
			case 0 : gap = 1; break;
			case 9 :
			case 10: gap = 11; break;
			default: break;
		}
		switches = 0;  // no swaps
		top = numElements - gap;
		for( i=0; i<top; i++ ){
			id	hi, hj;
			locType a, b;

			j = i+gap;
			hi = [aList objectAt:i];	a = [hi key];
			hj = [aList objectAt:j];	b = [hj key];
			if( [self compare:a with:b] > 0 ) {
				switches++;    
				// swap with neighbor
				[aList replaceObjectAt:i with:hj];
				[aList replaceObjectAt:j with:hi];
			}
		}
    } while( switches || gap>1 );
}



- initObjectList
{
	id		obj;
	
   
	getrusage( RUSAGE_SELF, &stime );

	listOfObjects = [[List alloc] initCount:50];
	[BibTexParser initParsing];
	NXGetMemoryBuffer(memStream, &theBuffer, &length, &maxLength);
	obj = [[BibTeXObject alloc] initFromStream:memStream owner:self];

	while( obj ){
		[self insertObject:obj];
		obj = [[BibTeXObject alloc] initFromStream:memStream owner:self];
		//if( obj ) [obj parseSelf];
	}

#ifdef PLPL
	while( obj ){
		[listOfObjects addObject:obj];
		obj = [[BibTeXObject alloc] initFromStream:memStream owner:self];
	}
	[self sortArray:listOfObjects];
#endif

	getrusage( RUSAGE_SELF, &etime );
	sMSec = stime.ru_utime.tv_sec*1000 + stime.ru_utime.tv_usec/1000;
	eMSec = etime.ru_utime.tv_sec*1000 + etime.ru_utime.tv_usec/1000;
	uTime = eMSec - sMSec;
   
	sMSec = stime.ru_stime.tv_sec*1000 + stime.ru_stime.tv_usec/1000;
	eMSec = etime.ru_stime.tv_sec*1000 + etime.ru_stime.tv_usec/1000;
	sTime = eMSec - sMSec; 

	return self;
}


- (int)uTime
{
	return uTime;
}


- (int)sTime
{
	return sTime;
}


- initNewFile:(const char *)fName
{
		
	[super init];
	myName = NXCopyStringBuffer(fName);	
	memStream = NXOpenMemory( NULL, 0, NX_READWRITE );

	listOfObjects = [[List alloc] initCount:50];
	isNewFile = YES;
	return self;
	
}


- initFromFile:(const char *)fName
{

	if( fName==NULL) return [super free];	// should never happen
	
	[super init];
	myName = NXCopyStringBuffer(fName);

	// read the file

	memStream = NXMapFile(myName, NX_READWRITE);
	if( memStream==NULL ){
		free( myName );
		return [super free];
	}
	isNewFile = NO;
	[self initObjectList];
	lastSearchPos = -1;
	return self;
}


- changeFilenameTo:(const char *)newName
{
	free( myName );
	myName = NXCopyStringBuffer(newName);
	isDirty = YES;
	return self;
	
}


// save  returns -1 on error, NX_OKTAG otherwise

- (int) save
{
	int	i;
	id	preferences;
	locType	r;	// (text)range of an object
	FILE	*fp;
	
	fp = fopen( myName, "w+" );
	
	if( fp==NULL ){
		return -1;
	}

	if( STRING<=0 ){
		preferences = [Preferences new];
		STRING = [preferences indexOfEntryname:"STRING"];
		PREAMBLE = [preferences indexOfEntryname:"PREAMBLE"];
	}

	NXGetMemoryBuffer(memStream, &theBuffer, &length, &maxLength);
	
	// first write out all PREAMBLEs
	for(i=0; i< [listOfObjects count]; i++ ){
		id	obj = [listOfObjects objectAt:i];

		if( [obj entryType]==PREAMBLE ) OUTPUT(obj);
	}

	// first write out all STRINGs
	for(i=0; i< [listOfObjects count]; i++ ){
		id	obj = [listOfObjects objectAt:i];

		if( [obj entryType]==STRING ) OUTPUT(obj);
	}

	for(i=0; i< [listOfObjects count]; i++ ){
		id	obj = [listOfObjects objectAt:i];

		if( [obj entryType]==STRING ) continue;
		if( [obj entryType]==PREAMBLE ) continue;
		// crossref should be the last to print
		NEWLINE;
		OUTPUT(obj);
	}

	NEWLINE;
	NEWLINE;

	fclose( fp );
	isDirty = NO;
	isNewFile = NO;
	
	return NX_OKTAG;
}


- (char *)fullPath
{
	return myName;
}


- (char *)fileName
{
	return rindex( myName, '/' )+1;
}

- (int) fileSize
{
	NXGetMemoryBuffer(memStream, &theBuffer, &length, &maxLength);
	return length;
}


- (int) entries
{
	return [listOfObjects count];
}


- objectAt:(unsigned int)index
{
	return [listOfObjects objectAt:index];
}


- (NXStream	*)stream
{
	return memStream;
}


- objectAtCharpos:(int)pos andSelectInMatrix:matrix
{
	int	i;
	locType	range;
	
	for( i=0; i<[listOfObjects count]; i++ ){
		range = [[listOfObjects objectAt:i] range];
		if( (range.start<=pos) && (pos<(range.start + range.length)) ){
			if( matrix ){
				[matrix selectCellAt:i :0];
				[matrix scrollCellToVisible:i :0];
			}
			return [listOfObjects objectAt:i];
		}
	}
	return NULL;
}


- showFile
{
	// only for testing :
	if( theText==NULL )		
		[NXApp loadNibSection:"BibText.nib" owner:self];

	NXFlush(memStream);
	NXSeek(memStream, 0, NX_FROMSTART);
	[theText readText:memStream];
	[[theText window] setTitleAsFilename:myName];
	[[theText window] orderFront:self];
	return self;
}


- (unsigned int)indexOf:obj
{
	return [listOfObjects indexOf:obj];
}


- findKey:(const char *)theKey
{
	int left=0;
	int	right=[listOfObjects count]-1;
	char keyBuffer[64];

	NXGetMemoryBuffer(memStream, &theBuffer, &length, &maxLength);	
	//   invariant : theKey must be between left and right
	while( left <= right ){
		int	cmp;
		int	i=(left + right)/2;
		id	theObj = [listOfObjects objectAt:i];
		locType keyRange = [theObj key];

		[theObj copy:keyRange.start :keyRange.length toBuffer:keyBuffer];
		cmp = strcasecmp( keyBuffer, theKey );
		if( cmp == 0 ) return theObj;
		if( cmp > 0 ) right = i-1;   else left = i+1;
	}
	//   invariant : theKey must be between left and right
	return nil;	// not found
}


- MYfindKey:(const char *)theKey
{
	int left=0;
	int	right=[listOfObjects count]-1;

	NXGetMemoryBuffer(memStream, &theBuffer, &length, &maxLength);	
	//   invariant : theKey must be between left and right
	while( left <= right ){
		int	j;
		char c;
		int	i=(left + right)/2;
		id	theObj = [listOfObjects objectAt:i];
		locType keyRange = [theObj key];
		char *objKey = &theBuffer[keyRange.start];
		int	len = keyRange.length;

		j = 0;
		while( c=theKey[j] ){
			if( objKey[j]==c ) j++;
			else break;
		}
		if( c=='\0' ){
			if(j==len) return theObj;	// found
			if(j<len) right = i-1; 
			else left = i+1;
		}
		else{	// c!= '\0'
			int cLower;
			if( NXIsUpper(c) ) c = NXToLower(c);
			cLower = objKey[j];
			if( NXIsUpper(cLower) ) cLower = NXToLower(cLower);
			if( c<cLower ) right = i-1;   else left = i+1;
		}

	}
	//   invariant : theKey must be between left and right
	return nil;	// not found
}


//	returns the list of all found objects

- findString:(const char *)pattern
		ignoreCase:(BOOL)ignoreCaseflag  tolerance:(int)tol
{
	unsigned int charVector[256];	// characteristic vector
	unsigned int bit[32], bit_0, endPos, wildCard;
	unsigned int findSet, findSet1;
	int i, j, patLen;
	int len;

	if( listOfFoundObjects==nil ){
		listOfFoundObjects = [[List alloc] init];
	}
	[listOfFoundObjects empty];
	
	for( i=0; i<32; i++ ) bit[i] = 1<<(32-1-i);
	bit_0 = bit[0];

	patLen = strlen( pattern );
	if( patLen>32 ) return nil;
	endPos = bit[ patLen-1 ];
  
	wildCard = 0;
	for( j=0; j<patLen; j++ ){	// are there wildcards (only '?' allowed)
		if( pattern[j]=='?' ) wildCard |= bit[j];
	}

	for( i=0; i<256; i++ ){	// define characteristic vector
		charVector[i] = wildCard;	// set is empty
	}

	for( j=0; j<patLen; j++ )
			charVector[ (unsigned int)pattern[j] ] |= bit[j];

	if( ignoreCaseflag==YES ){
		for( i='A'; i<='Z'; i++ ){
			unsigned int cv;
			cv = charVector[i] | charVector[i-'A'+'a'];
			charVector[i] = charVector[i-'A'+'a'] = cv;
		}
	}

	len = [self fileSize];
	findSet = findSet1 = 0;

	NXSeek( memStream, 0, NX_FROMSTART );
	lastSearchPos=0;

	for( j=1; j<=len; j++ ){
		unsigned int cv, theChar, h, w=findSet;

		theChar = NXGetc( memStream );
		lastSearchPos++;
		if( theChar==EOF ){
			lastSearchPos=0;
			return listOfFoundObjects;
		}
		cv = charVector[ theChar ];
		h = (findSet>>1) | bit_0;
		findSet = h & cv;

		if( tol )	// one error is allowed !!
			findSet1 = ((findSet1>>1) & cv) | h | (findSet>>1) | w;
		else 	findSet1 = findSet;

		if( findSet1 & endPos ){
			id	obj;
			//[theText setSel:lastSearchPos-patLen :lastSearchPos]; // test
			//[theText scrollSelToVisible];						// test
			obj = [self objectAtCharpos:lastSearchPos-patLen
					andSelectInMatrix:nil];
			if( obj ){
				if( obj!=[listOfFoundObjects lastObject] )
					[listOfFoundObjects addObject:obj];
			}
		}
    }
	// we will never reach this part
	lastSearchPos = -1;
	return listOfFoundObjects;
}



- free
{	
	[[listOfObjects freeObjects] free];

//NXGetMemoryBuffer(memStream, &theBuffer, &length, &maxLength);
//NXCloseMemory( memStream, NX_TRUNCATEBUFFER );
//vm_deallocate(task_self(), (vm_address_t)theBuffer, (vm_size_t )maxLength);

	free( myName );
	NXCloseMemory( memStream, NX_FREEBUFFER );
	if( theText )[[theText window] free];
	return [super free];
}


- setDirty
{
	isDirty = YES;
	return self;
}


- (BOOL)isDirty
{
	return isDirty;
}


- (BOOL)isNewFile
{
	return isNewFile;
}


- delete:obj
{
	[listOfObjects removeObject:obj];
	[obj free];
	return self;
}


- addObjects:objList	// list of matrixCells
{
	int		i;
	id		obj;
	int		lastPos;
	BOOL	sl;
	char *localBuffer;
	int	locBufferLength, locBufferMaxLength;

	if ( localStream==NULL ){
		localStream = NXOpenMemory( NULL, 0, NX_READWRITE );
	}
	NXSeek(localStream, 0, NX_FROMSTART);
	for( i=0; i<[objList count]; i++ ){
		NXPutc( localStream, '\n' );
		[[[objList objectAt:i] object] copyToStream:localStream];
		NXPutc( localStream, '\n' );
	}
	NXFlush( localStream );
	lastPos = NXTell( localStream );
	NXGetMemoryBuffer( localStream,
		&localBuffer, &locBufferLength, &locBufferMaxLength);
	NXSeek( memStream, 0, NX_FROMEND);	// append here
	NXWrite( memStream, localBuffer, lastPos );
	NXFlush( memStream );
	
	// parse the new string
	NXSeek( memStream, -lastPos, NX_FROMEND);	// go back
	obj = [[BibTeXObject alloc] initFromStream:memStream owner:self];
	
	sl = NO;
	while( obj ){
		[self insertObject:obj];
		[[NXApp delegate] update:obj for:self isNew:YES shiftLit:sl];
		sl = YES;
		obj = [[BibTeXObject alloc] initFromStream:memStream owner:self];
	}

	isDirty = YES;
	return self;
}


- appendNewText:(int)len :(char *)buffer
{
	id newObject, rtnObject;

	NXSeek( memStream, 0, NX_FROMEND);	// append here
	NXPrintf( memStream, "\n\n" );
	NXWrite( memStream, buffer, len );
	NXFlush( memStream );
	
	// parse the new text
	NXSeek( memStream, -len-1, NX_FROMEND);	// go back
	
	rtnObject = newObject
			  = [[BibTeXObject alloc] initFromStream:memStream owner:self];
	
	while( newObject ){
		[self insertObject:newObject];
		newObject = [[BibTeXObject alloc] initFromStream:memStream owner:self];
		if( newObject ) rtnObject = newObject;
	}
	isDirty = YES;
	return rtnObject;
}


- replace:(int)len :(char *)buffer forObject:obj
{
	locType		newRange;

	NXSeek( memStream, 0, NX_FROMEND);	// append here
	newRange.start =  NXTell( memStream );	// new position
	NXWrite( memStream, buffer, len );
	NXFlush( memStream );

	newRange.length = len;
	[obj setRange:newRange];
	[obj parseSelf];	// parse again

	isDirty = YES;
	return self;
}


@end
