

#import "Controller.h"
#import "DragView.h"
#import "BibTeXObject.h"
#import "MyCell.h"
#import "Shelf.h"
#import "Preferences.h"
#import "BufferClass.h"


static const char * dragTypes[3];
NXAtom	MyListPboardType;

id		entryImage = nil;
id		entryImageH;
id		entryImageMulti;
id		entryImageMultiH;
id		fileImage = nil;
id		fileImageH;
id		fileImageMulti;
id		fileImageMultiH;
id		alertImage;
id		closedBag;
id		smallImage;


#define ENTRY		1
#define BIBOBJECT	2
#define BIBFILE		3
#define TXTOBJECT	4

#define DRAGVIEWTAG	7


static id globalBuffer;


@implementation DragView


- initialize
{
	if( entryImage==nil ){	// set class variables

		MyListPboardType = NXUniqueString( "My List Pboard Type" );

		entryImage = [NXImage findImageNamed:"text"];
		entryImageH = [NXImage findImageNamed:"text"];
		entryImageMulti = [NXImage findImageNamed:"multiEntries"];
		entryImageMultiH = [NXImage findImageNamed:"multiEntries"];

		fileImage = [NXImage findImageNamed:"bibA"];
		fileImageH = [NXImage findImageNamed:"bibB"];
		fileImageMulti = [NXImage findImageNamed:"multiBook"];
		fileImageMultiH = [NXImage findImageNamed:"multiBook"];
		
		smallImage = [NXImage findImageNamed:"smallIcon"];
		alertImage = [NXImage findImageNamed:"bibAlert"];
		closedBag = [NXImage findImageNamed:"closedBag"];

		globalBuffer = [[BufferClass alloc] initBufferSize:128];
	}
	theObject = [[List alloc] init];
	theCell = [[MyCell alloc] initTextCell:"empty"];
	preferences = [Preferences new];
	return self;
}


- initFrame:(const NXRect *)frameRect
{ 
	[super initFrame:frameRect]; 
	isSelected = NO;
	return [self initialize];
}


- setShelf:theShelf
{
	shelf = theShelf;
	return self;
}


- awake	// called when unarchived
{ 
	return [self initialize];
}


- free
{
	[self clear];
	[theObject free];
	[theCell free];
	if( fileName ) free( fileName );
	return [super free];
}


- display
{
	char locBuf[64];
	char *fName;

	id	titleField = [[self superview] findViewWithTag:6];

	if( titleField==nil ) return [super display];

	if( theType == BIBFILE ){
		fName = rindex( fileName, '/' );
		if( fName )
			 [titleField  setStringValue:fName+1];
		else [titleField  setStringValue:fileName];
		return [super display];
	}

	if( theImage==nil ) [titleField setStringValue:NULL];
	else
	if( [theObject count]>1 ){
		sprintf( locBuf, "%d selected", [theObject count] );
		[titleField  setStringValue:locBuf];
	}
	else
		if( theType ){
			[titleField  setStringValue:
					[[theObject objectAt:0] stringValue]];
		}
	[super display];
	return self;
}


- (void) setFilename:(const char *)fName
{
	if( fileName ) free( fileName );
	fileName = NXCopyStringBuffer( fName );
	[theCell setObject:nil inFile:nil];
	theImage = anImage = fileImage;
	anImageH = fileImageH;
	anImageMulti = fileImageMulti;
	anImageMultiH = fileImageMultiH;
	theType = BIBFILE;
	[theObject empty];
	theImage = closedBag;
	if( access( fName, R_OK)<0 ){	// cannot read from file
		theImage = alertImage;
	}

	[self display];
}


- (char *) fileName
{
	if( theType==BIBFILE) return fileName;
	if( (theType==BIBOBJECT) && ([theObject count]==1) )
		return [[[theObject objectAt:0] fileObject] fullPath];

	return NULL;
}


- (void) setList:(List *)aList	// aList is a list of NXBrowsercells
{
	int	newType;

	if( aList!=theObject ){
		[theObject empty];
		[theObject appendList:aList];
	}
	if( aList==nil ){
		[self clear];
		return;
	}
	if( [[aList objectAt:0] object] ){	// bib_entry exist
		newType = ENTRY;
	}
	else  newType = BIBOBJECT;

	if( newType!=theType ){
	   switch( newType ){
		  case ENTRY : anImage = entryImage;
					 anImageH = entryImageH;
					 anImageMulti = entryImageMulti;
					 anImageMultiH = entryImageMultiH;
					 break;
		  case BIBOBJECT: anImage = fileImage;
					 anImageH = fileImageH;
					 anImageMulti = fileImageMulti;
					 anImageMultiH = fileImageMultiH;
					 break;
		  default: break;
	  }	
	  theType = newType;
	}

	if( [theObject count]==1 ) theImage = anImage;
	else theImage = anImageMulti;
	[self display];
}


- getList	// returns a list of NXBrowsercells
{
	return theObject;
}


- (unsigned int) count
{
	return [theObject count];
}


- (void) setTextObject:(Text *)aText
{
	[theObject empty];
	[theObject addObject:aText];
	theType = TXTOBJECT;
	theImage = smallImage;
	anImage = theImage;
	anImageH = theImage;
	anImageMulti = theImage;
	anImageMultiH = theImage;
	[self display];

}


- (void )setFileObject:(BibliographicFile *)aFile
{
	// use a NXBrowserCell for compatibility
	[theCell setObject:nil inFile:aFile];
	[theCell setStringValue:[aFile fileName]];
	[theObject empty];
	[theObject addObject:theCell];
	[self setList:theObject];
}


- (int)tag
{
	return DRAGVIEWTAG;
}


- drawSelf:(const NXRect *)rects :(int)rectCount
{
	if( isSelected ) PSsetgray( NX_WHITE );
	else	PSsetgray( NX_LTGRAY );	
	NXRectFill( &bounds );
	if( theImage ){
		[theImage composite:NX_SOVER toPoint:&bounds.origin];	// was &r.org
	}
	return self;
}


- (void) clear
{
	theImage = nil;
	[theObject empty];
	[self display];
}


- unregister
{
	if( theType!=BIBOBJECT ) return self;
	theType = BIBFILE;
	theImage = closedBag;
	[self unregisterDraggedTypes];
	[self display];
	return self;
}


//vvvvvvvvvvvvvvvvv		receiving  dragging

// includesType() returns YES if type is included within types.

static BOOL includesType (const NXAtom *types, NXAtom type)
{
    if (types)
		while (*types){
			//fprintf( stderr, "%s\n", *types );
			if (*types++ == type) return YES;
		}
    return NO;
}


- (NXDragOperation)draggingEntered:(id <NXDraggingInfo>)sender
{
	NXDragOperation rtn = [self draggingUpdated:sender];
	if( (rtn==NX_DragOperationGeneric) || (rtn==NX_DragOperationCopy) ){
		// change the Image
		theImage = anImageH;
		[self display];
	}
    return rtn;
}


- (NXDragOperation)draggingUpdated:(id <NXDraggingInfo>)sender
{
	Pasteboard *pboard = [sender draggingPasteboard];
	char *data, *lastDot, *file, *tab;
	int length;
	
	if( [sender draggingSource]==self ){
		fprintf( stderr, "it's me\n" );
		return NX_DragOperationNone;
	}

    if( ([sender draggingSourceOperationMask] & NX_DragOperationGeneric) 
	  ||([sender draggingSourceOperationMask] & NX_DragOperationCopy) ){

		if( includesType([pboard types], MyListPboardType) ){		
			//fprintf( stderr, "MyListPboardType in fileDragView\n" );
			return NX_DragOperationCopy;
		}
		else
		if( includesType([pboard types], NXAsciiPboardType) ){		
			//fprintf( stderr, "NXAsciiPboardType in fileDragView\n" );
			return NX_DragOperationCopy;
		}
		else
		if( includesType([pboard types], NXFilenamePboardType) ){
    		[pboard readType:NXFilenamePboardType data:&data length:&length];
			// is it a list of files ?
			file = data;
			while (file) {
				if( tab = strchr(file, '\t') ) { *tab = '\0'; }
				lastDot = rindex( file, '.' );
				if( lastDot ){ 
					if( strcmp( lastDot, ".bib" ) ) {	// only .bib allowed
						return NX_DragOperationNone;
					}
				}
				else return NX_DragOperationNone;	// no suffix -> invalid
				file = (tab) ? ++tab : NULL;
			}
			return NX_DragOperationGeneric;
		}
    }
	fprintf( stderr, "komisch\n" );
    return NX_DragOperationNone;	    
}


- draggingExited:sender
{
    //fprintf( stderr, "DragView draggingExited\n" );
	theImage = anImage;
	[self display];
    return self;
}


- (BOOL)performDragOperation:(id <NXDraggingInfo>)sender
{
	char *data, *file, *tab;
	id	*aList;
	int length;
	BOOL	rtn = NO;

	Pasteboard *pboard = [sender draggingPasteboard];

	if (includesType([pboard types], MyListPboardType)) {
		[pboard readType:MyListPboardType data:&data  length:&length];
		aList = (id *)data;

		// theObject is a list of one NXBrowsercell
		[[[theObject objectAt:0] fileObject] addObjects:*aList];

		rtn = YES;
	}

	if (includesType([pboard types], NXAsciiPboardType)) {
		[pboard readType:NXAsciiPboardType data:&data  length:&length];

		// theObject is a list of one NXBrowsercell
		[[[theObject objectAt:0] fileObject]
			appendNewText:length :data];
	
		 [pboard deallocatePasteboardData:data length:length];
		rtn = YES;
	}

	if(0)if( includesType([pboard types], NXFilenamePboardType) ){
		[pboard readType:NXFilenamePboardType data:&data length:&length];

		file = data;
		while (file) {
			if (tab = strchr(file, '\t')) {
				*tab = '\0';
			}
			//[self openFile:file];
			file = (tab) ? ++tab : NULL;
		}
	}

	// change the Image
	theImage = anImage;
	[self display];

	return rtn;
}


//^^^^^^^^^^^^^^^^^		end of receiving  dragging

//vvvvvvvvvvvv		active  dragging

- (NXDragOperation)draggingSourceOperationMaskForLocal:(BOOL)isLocal
{
	return NX_DragOperationGeneric | NX_DragOperationCopy;
}


- mouseDown:(NXEvent *)e
{
    NXPoint startPoint = e->location;
	int		nrOfClicks = e->data.mouse.click;

	savedEvent = *e;
    oldMask = [[self window] addToEventMask:NX_MOUSEDRAGGEDMASK];
    [self convertPoint:&startPoint fromView:nil];

	if( theType == TXTOBJECT ){
		//fprintf( stderr, "mouse down for text\n" );
		return self;
	}

	if( nrOfClicks==1111 ){
		isSelected = YES;
		[self display];
		return self;
	}

	if( isSelected ){
		isSelected = NO;
		[self display];
	}

	if( nrOfClicks==1 ){
		char *theName;
		isSelected = NO;	//YES;
		if( shelf==nil ) return self;

		if( theType == BIBFILE ) theName = fileName;
		else theName = [[[theObject objectAt:0] fileObject] fullPath];
		[[NXApp delegate] shelfText:theName];
		return self;
	}
	if( nrOfClicks==2 ){
		if( theType == BIBFILE ){
			int	okFlag;
			[NXApp openFile:fileName ok:&okFlag];
		}
		[[NXApp delegate] selectFile:[[theObject objectAt:0] fileObject]];
	}
	return self;
}



- mouseUp:(NXEvent *)e
{
	[[self window] setEventMask:oldMask];
    return self;
}



- mouseDragged:(NXEvent *)theEvent
{
	int	i;
	NXPoint	point = theEvent->location;
	NXPoint	origin = {0.0, 0.0};
	char	fName[128];
	Pasteboard *pboard;

	if( !theObject ) return self;

    [NXApp preventWindowOrdering];
	
	[self convertPoint:&point fromView:nil];
	//[super mouseDown:&savedEvent];
	
/* You always use the NXDragPboard pasteboard when dragging. */
	pboard = [Pasteboard newName:NXDragPboard];

	switch( theType ){
	
	case 0:	return self;

	case BIBOBJECT:
		/*  place the name of the file on the pasteboard. */
			[pboard declareTypes:&NXFilenamePboardType num:1 owner:nil];
			// theObject is a list of one NXBrowsercell
			// first we write the names into a stream
			if( !dragStream )
				dragStream = NXOpenMemory(NULL, 0, NX_READWRITE);
			NXSeek(dragStream, 0, NX_FROMSTART);

			for( i=0; i<[theObject count]; i++ ){
				NXPrintf( dragStream, "%s",
					[[[theObject objectAt:i] fileObject] fullPath] );
				if( i==[theObject count]-1 )  NXPutc(dragStream, '\0');
				else  NXPutc(dragStream, '\t');
			}
			[pboard writeType:NXFilenamePboardType fromStream:dragStream];
			NXCloseMemory( dragStream, NX_FREEBUFFER );
			dragStream = NULL;
			break;
	
	case ENTRY:	{
			FILE	*fileStream;
			int	i;

			sprintf( fName, "%s/citeFile", [preferences tempDirectory] );
			if( (fileStream = fopen( fName, "w+" ))==NULL ){
				fprintf( stderr, "cannot open %s\n", fName );
				return self;
			}
			for( i=0; i<[theObject count]; i++ ){
				sprintf( [globalBuffer buffer], "\\cite{%s}  ",
						[[theObject objectAt:i] stringValue] );
				fwrite( [globalBuffer buffer], 1,
					strlen( [globalBuffer buffer] ), fileStream );
			}
			fclose( fileStream );

		/*  place the filename on the pasteboard. */
			[pboard declareTypes:&NXFilenamePboardType num:1 owner:nil];
			sprintf( [globalBuffer buffer], fName );
			[pboard writeType:NXFilenamePboardType
				data:[globalBuffer buffer]
				length:strlen( [globalBuffer buffer] )+1];

		/*  place also the list of entries on the pasteboard. */
			[pboard addTypes:&MyListPboardType num:1 owner:self];
			[pboard writeType:MyListPboardType
				data:(const char *)(&theObject) length:sizeof(id) ];
			break;
			}

	case BIBFILE:
		/*  place the name of the file on the pasteboard. */
			[pboard declareTypes:&NXFilenamePboardType num:1 owner:nil];
			[pboard writeType:NXFilenamePboardType
				data:fileName length:strlen( fileName )+1];
			break;


	case TXTOBJECT:{
			int	length;
			id	theText = [theObject objectAt:0];
			NXSelPt start, end;
	
			if( theText==nil ) return self;
			[theText getSel:&start :&end];
			if( start.cp<0 ){	// text is not first responder
				return self;
			}
			length = end.cp - start.cp;
			if( length<=0 ){
				[theText selectAll:self];
			}
			/*  place the selected text on the pasteboard. */
			dragTypes[0] = NXAsciiPboardType;
			dragTypes[1] = NULL;
			[pboard declareTypes:dragTypes num:1 owner:nil];
			if( [theText writeSelectionToPasteboard:pboard types:dragTypes] ){
				break;
			}

			NXRunAlertPanel("Bibliography",
					"could not write to pasteboard",
					"  OK ",	// 1: default button
					NULL,		// 0: alternate
					NULL 		//-1: other
				);

			break;
		}

	default : fprintf( stderr, "unknown case %d\n", theType );
	 
	}


/* Now invoke dragImage: */
	[self dragImage:theImage
			at:&origin			offset:NULL		event:&savedEvent
			pasteboard:pboard	source:self		slideBack:NO];
	return self;
}


- draggedImage : (NXImage *)image 
	   endedAt : (NXPoint *)screenPoint
	 deposited : (BOOL)didDeposit
{
	NXRect winRect;

	if( didDeposit ) return self;
	if( shelf==nil ) return self;	// shall not remove

	[[self window] getFrame:&winRect];

	if( NXPointInRect(screenPoint,&winRect) ){
		return self;
	}
	if(0)fprintf( stderr, "x=%d y=%d   %d\n",
		(int)screenPoint->x, (int)screenPoint->y, didDeposit );
	theType = 0;
	[shelf removeBoxWithIcon:self];
	return self;
}



- (BOOL)shouldDelayWindowOrderingForEvent:(NXEvent *)theEvent
{
 	// This is the other portion of the new appkit functionality
	// which allows you to drag a partially obscure object without
	// making the source window become main, and without making the
	// source application the current app.
    return YES;
}


- (BOOL)acceptsFirstMouse
{
	return YES;
}


//^^^^^^^^^^^^^^^^^		end of active  dragging

@end
