

#import "Controller.h"
#import "FindClass.h"
#import "MyCell.h"
#import "DragView.h"
#import "BibTexView.h"
#import "BibliographicFile.h"

	

@implementation FindClass


- init
{
	[super init];
	inspector = [BibTexView new];
	findPBoard = [Pasteboard newName:NXFindPboard];
	[latexButton setEnabled:NO];
	return self;
}



// aList is a list of matrixcells of class MyCell

- (void)deleteEntries:(List *)aList
{
	int	i, j, rows, cols;
	char	buffer[64];

	if( [listOfFoundPatterns count]==0 ) return;

	if( findMatrix==nil ) return;

	for( j=0; j<[aList count]; j++ ){
		id	delObj = [[aList objectAt:j] object];	// entry_object to delete
		[findMatrix getNumRows:&rows numCols:&cols];
		for( i=rows-1; i>=0; i-- ){
			id	anObj = [[findMatrix cellAt:i :0] object];
			if( anObj==delObj ){	// both are of class BibTexObject
				[findMatrix removeRowAt:i andFree:NO];
				[listOfFoundPatterns removeObject:delObj];
				if( lastFindObject==delObj ) lastFindObject=nil;
				continue;
			}
		}
	}
	[findMatrix sizeToCells];
	[listOfSelectedPatterns free];
	listOfSelectedPatterns = [findMatrix getSelectedCells:nil];
	sprintf( buffer, "%d selected", [listOfSelectedPatterns count] );
	[selectedIconText setStringValue:buffer];

	if( [listOfSelectedPatterns count]>0 ){
		[latexButton setEnabled:YES];
		[findIcon setList:listOfSelectedPatterns];
	}
	else{
		[latexButton setEnabled:NO];
		[findIcon setList:nil];
	}
	[[findMatrix window] display];
}


- find:sender
{
	NXRect frame = {{0.0, 0.0},{100.0, 100.0}};

	if( !findMatrix ){
		findMatrix = [[Matrix alloc] initFrame:&frame
					mode:NX_LISTMODE
					cellClass:[MyCell class]
					numRows:0 numCols:1];
		[findMatrix setCellBackgroundGray:NX_LTGRAY];
		[findMatrix setBackgroundGray:NX_LTGRAY];
		[findMatrix setAutoscroll:YES];
		[findScroller setBackgroundGray:NX_LTGRAY];
		[[findScroller setDocView:findMatrix] free];
		[[findMatrix window]setFrameUsingName:"FindPanel"];
		[[findMatrix window] setFrameAutosaveName:"FindPanel"];
		[latexButton setEnabled:NO];
	}

	[[findScroller window] makeKeyAndOrderFront:self];
	return self;
}


//	aList contains entry-objects
- copy:(List *)aList toFile:(const char *)aFile	// no longer used
{
	int	i;

	if( !findStream ){
		findStream = NXOpenMemory(NULL, 0, NX_READWRITE);
	}
	NXSeek(findStream, 0, NX_FROMSTART);

	for( i=0; i<[aList count]; i++ ){
		NXPutc( findStream, '\n');
		[[aList objectAt:i] copyToStream:findStream];
		NXPutc( findStream, '\n');
	}

	NXSaveToFile( findStream, aFile );
	NXCloseMemory( findStream, NX_FREEBUFFER );
	findStream = NULL;
	return self;
}


- (void)showFindList
{
	int	i;
	NXRect	frame = {{0.0, 0.0},{100.0, 100.0}};
	NXSize	theSize;
	char	keyBuffer[128];
	char	buffer[512];

	if( findMatrix ){
		[findMatrix renewRows:0 cols:0];
		[findMatrix sizeToCells];
	}
	else{
		findMatrix = [[Matrix alloc] initFrame:&frame
					mode:NX_LISTMODE
					cellClass:[MyCell class]
					numRows:0 numCols:1];
		[findMatrix setCellBackgroundGray:NX_LTGRAY];
		[findMatrix setBackgroundGray:NX_LTGRAY];
		[findMatrix setAutoscroll:YES];
		[findScroller setBackgroundGray:NX_LTGRAY];
		[[findScroller setDocView:findMatrix] free];
	}

	[findScroller getFrame:&frame];
	[findScroller getContentSize:&frame.size];
	
	[findMatrix setAction:@selector(singleClickInFindMatrix:)];
	[findMatrix getCellSize:&theSize];
	[findMatrix setAutosizeCells:YES];	// tut nicht
	theSize.width = frame.W;
	[findMatrix setCellSize:&theSize];
	
	for( i=0; i<[listOfFoundPatterns count]; i++ ){
		id newCell;
		id	obj = [listOfFoundPatterns objectAt:i];
		id	fileObj = [obj fileObject];

		if( i<0 ){	// TEST
			[findMatrix addRow];
			newCell = [findMatrix cellAt:i :0];
			[newCell setFont:[Font newFont:"Courier" // tut nicht
							size:14.0
							style:0
							matrix:NX_IDENTITYMATRIX]];
			sprintf( buffer, "    file : %s", [fileObj fileName] );
			[newCell setStringValue:buffer];
			[newCell setObject:obj inFile:fileObj];
		}
		
		[findMatrix addRow];
		newCell = [findMatrix cellAt:i :0];
		[newCell setLeaf:YES];	// is NXBrowerCell !!
		[obj copyRange:[obj key] toBuffer:keyBuffer];
		sprintf( buffer, "%s : %s", [fileObj fileName], keyBuffer );
		[newCell setStringValue:buffer];
		[newCell setObject:obj inFile:fileObj];
	}
	
	[findMatrix selectCellAt:-1 :-1];
	[[[findMatrix sizeToCells] window] display];
	
	sprintf( buffer, "%d found", [listOfFoundPatterns count] );
	[findIconText setStringValue:buffer];
	[selectedIconText setStringValue:""];
	
	[findIcon setList:nil];
	[findIcon display];

	[latexButton setEnabled:NO];
}


- (void)setFindText
{
	char *data;
	int length;
	const NXAtom *array;

	array = [findPBoard types];
	if( *array ){
    	if( [findPBoard readType:*array data:&data length:&length] ){
			[searchTextField setStringValue:data];
		}
	}
	[searchTextField selectText:self];
}


//	objects in selList are of class MyCell
- (void)findStringIn:(List *)selList
{
	int	index;
	id	fileObj;
	id	rtnList;

	lastFindObject = lastFindFile = nil;
	// write to NXFindPboard
	[findPBoard declareTypes:&NXAsciiPboardType num:1 owner:self];
	[findPBoard writeType:NXAsciiPboardType data:[searchTextField stringValue] 
			length:strlen([searchTextField stringValue])];

	if( !listOfFoundPatterns )
			listOfFoundPatterns = [[List alloc] init];
	if( ![addButton state] )[listOfFoundPatterns empty];
	
	for( index=0; index<[selList count]; index++ ){
		fileObj = [[selList objectAt:index] fileObject];
		rtnList = [fileObj findString:[searchTextField stringValue]
			ignoreCase:[ignoreCase state] tolerance:[allowErrors state]];
		[listOfFoundPatterns appendList:rtnList];
	
		// check for multiple entries
		// (not yet implemented)
	}

	[self showFindList];
}


- singleClickInFindMatrix:sender
{
	id	cell = [sender selectedCell];
	char	buffer[64];
	
	
	[[NXApp delegate] P:self];
	[listOfSelectedPatterns free];
	listOfSelectedPatterns = [sender getSelectedCells:nil];
	if( [listOfSelectedPatterns count]==1 ){
		[inspector setEntry:(lastFindObject = [cell object])
				inFile:(lastFindFile = [cell fileObject]) ];
	}
	else {
		[inspector clear:self];
		lastFindObject = lastFindFile = nil;
	}
	[findIcon setList:listOfSelectedPatterns];
	sprintf( buffer, "%d selected", [listOfSelectedPatterns count] );
	[selectedIconText setStringValue:buffer];

	[latexButton setEnabled:YES];
	[findIcon display];
	[[NXApp delegate] V:self];
	return self;
}


- (void)deleteFile:(BibliographicFile *)theFile
{
	int	i;
	int	row, col;
	BOOL	needsRedraw = NO;

	if( !findMatrix ) return;
	[findMatrix  getNumRows:&row numCols:&col];
	for( i=row-1; i>=0; i-- ){
		id	cell = [findMatrix cellAt:i :0];
		if( [cell fileObject] == theFile ){
			if( theFile==lastFindFile ){
				lastFindObject = lastFindFile = nil;
			}
			[listOfFoundPatterns removeObject:[cell object]];
			needsRedraw = YES;
		}
	}
	if( needsRedraw==YES ) [self showFindList];
}


- prepareLaTeX
{
	//[[NXApp delegate] LaTeXFromMatrix:findMatrix];
	[[NXApp delegate] LaTeXFromList:[findIcon getList]];
	return self;
}


- LaTeX:sender
{
	//[self prepareLaTeX];
	[[NXApp delegate] LaTeXFromList:[findIcon getList]];
	return self;
}


- windowDidBecomeKey:sender
{
	if( [[NXApp delegate] semaphore:self] <= 0 ) return self;

	[[NXApp delegate] P:self];

	if( lastFindObject==nil ) [inspector clear:self];
	else
	[inspector setEntry:lastFindObject inFile:lastFindFile ];

	[[NXApp delegate] V:self];
	[findButton setIconPosition:NX_ICONRIGHT];
	[sender flushWindow];
	return self;
}


- windowDidResignKey:sender
{
	[sender endEditingFor:nil];
	[self setFindText];
	[findButton setIconPosition:NX_TITLEONLY];
	[findButton display];
	return self;
}


//	returns the new state of the menuCell
	// YES = enabled
	//	NO = disabled

- (BOOL)updateMenuFor:menuCell
{
	BOOL	isEnabled = [menuCell isEnabled];
	BOOL	newState=NO;	// .. and the compiler keeps silent
	int	row = [findMatrix selectedRow];

	switch( [menuCell tag] ){
		case CUT		: newState = YES; break;
		case COPY		: newState = YES; break;
		case PASTE		: newState = YES; break;
		case SELECTALL	: newState = YES; break;;
		case ENTERSELECTION : newState = NO; break;

		case PRVWINDOW	: newState = YES; break;
		case RUNLATEX	: if( findMatrix==nil )	newState = NO;
							else newState = (row>=0); break;
		case PRINT		: newState = isEnabled;
								break;
		default : newState = [[NXApp delegate]updateMenuFor:menuCell];
				  break;
		//newState = isEnabled;	// do not change
		}

	return newState;
}


@end
